package sketch;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

import dna.Parser;
import shared.Shared;
import shared.Tools;
import structures.Heap;
import tax.TaxNode;

public class SketchSearcher extends SketchObject {
	
	public SketchSearcher(){
		
	}

	public boolean parse(String arg, String a, String b){
		if(Parser.isJavaFlag(arg)){
			//do nothing
		}else if(parseSketchFlags(arg, a, b)){
			//Do nothing
		}else if(defaultParams.parse(arg, a, b)){
			//Do nothing
		}else if(a.equals("verbose")){
			verbose=Tools.parseBoolean(b);
		}else if(a.equals("ref")){
			addFiles(b, refFiles);
		}else if(a.equals("threads") || a.equals("sketchthreads") || a.equals("t")){
			threads=Integer.parseInt(b);
		}
		
		else if(a.equals("index") || a.equals("makeindex")){
			if(b!=null && "auto".equalsIgnoreCase(b)){
				autoIndex=true;
				makeIndex=true;
			}else{
				autoIndex=false;
				makeIndex=Tools.parseBoolean(b);
			}
		}else if(a.equals("indexsize") || a.equals("indexlimit")){
			SketchIndex.indexLimit=Integer.parseInt(b);
		}
		
		else if(b==null && arg.indexOf('=')<0){//if(new File(arg).exists())
			refFiles.add(arg);
		}else{
			return false;
		}
		return true;
	}

	public boolean compare(ArrayList<Sketch> sketches, StringBuilder sb, DisplayParams params){
		ConcurrentHashMap<Integer, Comparison> map=new ConcurrentHashMap<Integer, Comparison>();

		@SuppressWarnings("unchecked")
		ArrayList<Comparison>[] alca=new ArrayList[sketches.size()];

		boolean success=true;
		final int[] buffer=Sketch.makeBuffer();
		AtomicInteger fakeID=new AtomicInteger(minFakeID);
		for(int i=0; i<sketches.size(); i++){
			fakeID.set(minFakeID);
			Sketch a=sketches.get(i);
			if(index!=null){
				processUsingIndex(a, buffer, fakeID, map, params);
			}else if(threads<2){
				for(Sketch b : refSketches){
					processPair(a, b, buffer, fakeID, map, params);
				}
			}else{
				ArrayList<CompareThread> alct=new ArrayList<CompareThread>(threads);
				for(int t=0; t<threads; t++){
					alct.add(new CompareThread(a, t, threads, fakeID, map, params));
				}
				for(CompareThread ct : alct){ct.start();}
				for(CompareThread ct : alct){
					
					//Wait until this thread has terminated
					while(ct.getState()!=Thread.State.TERMINATED){
						try {
							//Attempt a join operation
							ct.join();
						} catch (InterruptedException e) {
							e.printStackTrace();
						}
					}
				}
				alct=null;
			}
			
			ArrayList<Comparison> al=alca[i]=new ArrayList<Comparison>(map.size());
			for(Entry<Integer, Comparison> e : map.entrySet()){
				al.add(e.getValue());
			}
			Shared.sort(al);
			Collections.reverse(al);
			while(al.size()>params.maxRecords){
				al.remove(al.size()-1);
			}
			map.clear();
		}
		
		for(int i=0; i<alca.length; i++){
			Sketch s=sketches.get(i);
			ArrayList<Comparison> al=alca[i];
			sb.append(toText(al, s, params.format, params.printTax, params.printOriginalName));
		}
		return success;
	}
	
	private class CompareThread extends Thread {
		
		CompareThread(Sketch a_, int pid_, int incr_, AtomicInteger fakeID_, ConcurrentHashMap<Integer, Comparison> map_, DisplayParams params_){
			a=a_;
			pid=pid_;
			incr=incr_;
			fakeID=fakeID_;
			map=map_;
			params=params_;
		}
		
		public void run(){
			if(index!=null){
				processUsingIndex(a, buffer, fakeID, map, params);
			}else{
				for(int i=pid; i<refSketches.size(); i+=incr){
					Sketch b=refSketches.get(i);
					processPair(a, b, buffer, fakeID, map, params);
				}
			}
		}
		
		final AtomicInteger fakeID;
		final ConcurrentHashMap<Integer, Comparison> map;
		final int[] buffer=Sketch.makeBuffer();
		final int incr;
		final int pid;
		final Sketch a;
		final DisplayParams params;
		
	}
	
	void processUsingIndex(Sketch a, int[] buffer, AtomicInteger fakeID, ConcurrentHashMap<Integer, Comparison> map, DisplayParams params){
//		Timer t=new Timer();
//		t.start("Began query.");
		ArrayList<Sketch> hits=index.getSketches(a, params.minCount);
//		t.stop("Got "+hits.size()+" hits.");
//		t.start();
//		System.err.println("hits: "+hits.size());
		if(hits==null || hits.isEmpty()){return;}
		comparisons.getAndAdd(hits.size());
		for(Sketch b : hits){
			processPair(a, b, buffer, fakeID, map, params);
		}
//		t.stop("Processed pairs.");
	}
	
//	private void writeResults(ArrayList<Comparison> al, Sketch s, StringBuilder sb){
//		sb.append("\nResults for "+s.name()+":\n\n");
//		
//		ArrayList<TaxNode> tnl=new ArrayList<TaxNode>();
//		for(Comparison c : al){
//			formatComparison(c, format, sb, printTax);
//		}
//	}
	
	boolean processPair(Sketch a, Sketch b, int[] buffer, AtomicInteger fakeID, ConcurrentHashMap<Integer, Comparison> map, DisplayParams params){
//		System.err.println("Comparing "+a.name()+" and "+b.name());
		Comparison c=compareOneToOne(a, b, buffer, params.minCount, params.minID, null);
		if(c==null){return false;}
		if(c.taxID<0){c.taxID=fakeID.getAndIncrement();}
		
		TaxNode tn=(taxtree==null ? null : taxtree.getNode(b.taxID));
		if(tn!=null){
			c.taxName=tn.name;
			if(tn.level<params.taxLevel){
				tn=taxtree.getNodeAtLevel(b.taxID, params.taxLevel);
			}
		}
		Integer key=(tn==null ? c.taxID : tn.id);

		Comparison old=map.get(key);
//		System.err.println("A. Old: "+(old==null ? 0 : old.hits)+", new: "+c.hits);
		if(old!=null && old.compareTo(c)>0){return false;}
		
		old=map.put(key, c);
		while(old!=null && old.compareTo(c)>0){
//			System.err.println("B. Old: "+(old==null ? 0 : old.hits)+", new: "+c.hits);
			c=old;
			old=map.put(key, c);
		}
		return true;
	}
	
	private static Comparison compareOneToOne(final Sketch a, final Sketch b, int[] buffer, int minCount, float minID, Heap<Comparison> heap){
//		assert(heap!=null); //Optional, for testing.
		final int matches=a.countMatches(b, buffer);
		assert(matches==buffer[0]);
		final int div=buffer[1];
		
		if(matches<minCount || matches/(float)div<minID){
//			System.err.print(".");
			return null;
		}
		if(heap!=null && !heap.hasRoom() && heap.peek().hits>matches){return null;}
		
//		System.err.print("*");
		Comparison c=new Comparison(buffer, b);
		if(heap==null || heap.add(c)){return c;}
		return null;
	}
	
	private void addFiles(String a, Collection<String> list){
		if(a==null){return;}
		File f=null;
		if(a.indexOf(',')>=0){f=new File(a);}
		if(f==null || f.exists()){
			list.add(a);
		}else{
			for(String s : a.split(",")){list.add(s);}
		}
	}
	
	public void loadReferences() {
		makeTool();
		refSketches=tool.loadSketches_MT(SketchObject.PER_TAXA, samplerate, refFiles);
//		System.err.println("Sketches: "+refSketches.get(0).name());
		if(makeIndex){
			index=new SketchIndex(refSketches);
			index.load();
		}
	}
	
	public void makeTool(){
		if(tool==null){
			tool=new SketchTool(size, defaultMinCount);
		}
	}
	
	public ArrayList<Sketch> loadSketchesFromString(String sketchString){
		return tool.loadSketchesFromString(sketchString);
	}
	
	/*--------------------------------------------------------------*/
	
	public SketchIndex index=null;
	public boolean autoIndex=true;
	public boolean makeIndex=true;
	
	public SketchTool tool=null;
	public ArrayList<Sketch> refSketches=new ArrayList<Sketch>();
	public ArrayList<String> refFiles=new ArrayList<String>();
	public int threads=Shared.threads();
	boolean verbose;
	boolean errorState=false;
	AtomicLong comparisons=new AtomicLong(0);
	
	public static final int defaultMinCount=3;
	
}
